@@ -1,11 +1,17 @@
 from dateutil.parser import parse as parse_date
-from django.db import models
+from django.conf import settings as django
+from django.db import models, transaction
+from django.dispatch import receiver
 from django.template.loader import render_to_string
 from django.urls import reverse
 from django.utils import html, text
+from django.utils import timezone
 from feedparser import parse as parse_feed
+from hashlib import md5
 from html2text import html2text
+from importlib import import_module
 from markdownx.models import MarkdownxField
+from pico import websub
 from pico.conf import settings
 from urllib.parse import urlparse
 from .query import PodcastQuerySet, EpisodeQuerySet, PostQuerySet
@@ -31,6 +37,12 @@ class Podcast(models.Model):
             os.path.splitext(filename)[-1]
         )
 
+    def upload_banner(self, filename):
+        return 'podcasts/%s/upload_banner%s' % (
+            self.slug,
+            os.path.splitext(filename)[-1]
+        )
+
     name = models.CharField(max_length=100)
     short_name = models.CharField(max_length=100, null=True, blank=True)
     slug = models.SlugField(max_length=30, unique=True)
@@ -42,6 +54,13 @@ class Podcast(models.Model):
         upload_to=upload_artwork
     )
 
+    banner = models.ImageField(
+        max_length=255,
+        upload_to=upload_banner,
+        null=True,
+        blank=True
+    )
+
     subtitle = models.CharField(max_length=255, null=True, blank=True)
     description = MarkdownxField()
 
@@ -126,90 +145,104 @@ class Podcast(models.Model):
         default='f4f0eb'
     )
 
+    subscription = models.ForeignKey(
+        'websub.Subscription',
+        on_delete=models.SET_NULL,
+        null=True,
+        editable=False
+    )
+
     def __str__(self):
         return self.name
 
+    def save(self, *args, **kwargs):
+        if not self.subscription:
+            self.subscription = websub.subscribe(self.rss_feed_url)
+
+        super().save(*args, **kwargs)
+
     def check_feed(self, episode_callback=None):
         feed_data = parse_feed(self.rss_feed_url)
         self.update_feed(feed_data, episode_callback)
 
     def update_feed(self, feed, episode_callback=None):
         for item in feed['items']:
-            season_number = item.get('itunes_season')
-            episode_number = item.get('itunes_episode')
-            episode_type = item.get('itunes_episodetype')
-            author = item.get('author')
-            season = None
+            with transaction.atomic():
+                season_number = item.get('itunes_season')
+                episode_number = item.get('itunes_episode')
+                episode_type = item.get('itunes_episodetype')
+                author = item.get('author')
+                season = None
 
-            if season_number:
-                season = self.seasons.filter(
-                    number=season_number
-                ).first()
-
-                if season is None:
-                    season = self.seasons.create(
+                if season_number:
+                    season = self.seasons.filter(
                         number=season_number
+                    ).select_for_update().first()
+
+                    if season is None:
+                        season = self.seasons.create(
+                            number=season_number
+                        )
+
+                image = item.get('image', {}).get('href', '')
+                if image:
+                    image = download(image)
+
+                    if self.artwork and not compare_image(image, self.artwork):
+                        image = None
+
+                description = item.get('summary')
+                summary = html.strip_tags(item.get('subtitle'))
+
+                for detail in item.get('content', []):
+                    if detail['type'] == 'text/html':
+                        description = detail['value']
+                        break
+
+                enclosure = None
+                for link in item.get('links', []):
+                    if link['rel'] == 'enclosure':
+                        enclosure = link['href']
+
+                date = parse_date(item['published'])
+                episode = self.episodes.filter(
+                    guid=item['id']
+                ).select_for_update().first()
+
+                if episode is None:
+                    episode = Episode(
+                        guid=item['id'],
+                        podcast=self
                     )
+                elif episode.artwork:
+                    if image and not compare_image(image, episode.artwork):
+                        image = None
 
-            image = item.get('image', {}).get('href', '')
-            if image:
-                image = download(image)
+                episode.title = item['title']
+                episode.published = date
+                episode.season = season
+                episode.number = episode_number or 0
+                episode.bonus = episode_type == 'bonus'
+                episode.trailer = episode_type == 'trailer'
 
-                if self.artwork and not compare_image(image, self.artwork):
-                    image = None
+                if image is not None:
+                    episode.artwork = image
 
-            description = item.get('summary')
-            summary = html.strip_tags(item.get('subtitle'))
+                episode.summary = summary
+                episode.feed_description = html2text(description)
+                episode.enclosure_url = enclosure
+                episode.save()
 
-            for detail in item.get('content', []):
-                if detail['type'] == 'text/html':
-                    description = detail['value']
-                    break
+                if author:
+                    for name in author.split(','):
+                        name_stripped = name.strip()
+                        host = Host.objects.filter(
+                            name__iexact=name_stripped
+                        ).first()
 
-            enclosure = None
-            for link in item.get('links', []):
-                if link['rel'] == 'enclosure':
-                    enclosure = link['href']
-
-            date = parse_date(item['published'])
-            episode = self.episodes.filter(
-                guid=item['id']
-            ).first()
-
-            if episode is None:
-                episode = Episode(
-                    guid=item['id'],
-                    podcast=self
-                )
-            elif episode.artwork:
-                if not compare_image(image, episode.artwork):
-                    image = None
-
-            episode.title = item['title']
-            episode.published = date
-            episode.season = season
-            episode.number = episode_number or 0
-            episode.bonus = episode_type == 'bonus'
-            episode.trailer = episode_type == 'trailer'
-
-            if image is not None:
-                episode.artwork = image
-
-            episode.summary = summary
-            episode.feed_description = html2text(description)
-            episode.enclosure_url = enclosure
-            episode.save()
-
-            if author:
-                for name in author.split(','):
-                    name_stripped = name.strip()
-                    host = Host.objects.filter(
-                        name__iexact=name_stripped
-                    ).first()
-
-                    if host is not None:
-                        episode.hosts.add(host)
-                        self.hosts.add(host)
+                        if host is not None:
+                            episode.hosts.add(host)
+                            self.hosts.add(host)
 
             if callable(episode_callback):
                 episode_callback(episode)
@@ -230,14 +263,66 @@ class Podcast(models.Model):
 
         return self.__apple_podcasts_id
 
+    def check_reviews(self, review_callback=None):
+        for link in self.subscription_links.iterator():
+            link.check_reviews(review_callback)
+
+    def get_menu_items(self):
+        yield {
+            'url': '/',
+            'text': 'Home'
+        }
+
+        for season in self.seasons.order_by('-number'):
+            yield {
+                'url': self.reverse(
+                    'season',
+                    args=(season.number,)
+                ),
+                'text': str(season)
+            }
+
+        if self.blog_posts.filter(
+            published__lte=timezone.now()
+        ).exists():
+            yield {
+                'url': self.reverse('blogpost_list'),
+                'text': 'Blog'
+            }
+
+        for page in self.pages.filter(menu_visible=True):
+            yield {
+                'url': self.reverse(
+                    'page_detail',
+                    args=(page.slug,)
+                ),
+                'text': page.menu_title or page.title,
+                'highlight': page.cta
+            }
+
+        if settings.CONTACT_FORM:
+            yield {
+                'url': '/contact/',
+                'text': 'Contact'
+            }
+
     def reverse(self, urlname, args=(), kwargs={}):
+        urlconf = 'pico.urls'
+
         if settings.DOMAINS_OR_SLUGS == 'slugs':
             if any(kwargs):
                 kwargs['podcast'] = self.slug
             else:
                 args = (self.slug,) + args
+        else:
+            urlconf = 'pico.podcasts.urls'
 
-        return reverse(urlname, args=args, kwargs=kwargs)
+        return reverse(
+            urlname,
+            args=args,
+            kwargs=kwargs,
+            urlconf=urlconf
+        )
 
     def build_absolute_uri(self, path=''):
         while path.startswith('/'):
@@ -286,6 +371,43 @@ class Directory(models.Model):
     def __str__(self):
         return self.name
 
+    def get_reviews(self, link):
+        module_name = text.slugify(self.name).replace('-', '_')
+        reviewer = django.REVIEW_DIRECTORIES.get(module_name)
+
+        if reviewer is None:
+            return
+
+        module, func = reviewer.rsplit('.', 1)
+        module = import_module(module)
+        func = getattr(module, func)
+
+        for review in func(link.url):
+            with transaction.atomic():
+                remote_id = review['id']
+                obj = link.podcast.reviews.filter(
+                    directory=self,
+                    remote_id=remote_id,
+                    country=review['country']
+                ).first()
+
+                if obj is None:
+                    obj = Review(
+                        podcast=link.podcast,
+                        directory=self,
+                        remote_id=remote_id,
+                        country=review['country']
+                    )
+
+                obj.title = review['title']
+                obj.body = review['body']
+                obj.author = review['author']
+                obj.published = review['published']
+                obj.rating = review['rating']
+                obj.save()
+
+            yield obj
+
     class Meta:
         ordering = ('ordering',)
         verbose_name_plural = 'directories'
@@ -309,6 +431,11 @@ class SubscriptionLink(models.Model):
     def __str__(self):
         return urlparse(self.url).netloc
 
+    def check_reviews(self, review_callback=None):
+        for review in self.directory.get_reviews(self):
+            if review_callback is not None and callable(review_callback):
+                review_callback(review)
+
     class Meta:
         ordering = ('directory__ordering',)
         unique_together = ('directory', 'podcast')
@@ -338,6 +465,13 @@ class Season(models.Model):
         blank=True
     )
 
+    banner = models.ImageField(
+        max_length=255,
+        upload_to=upload_artwork,
+        null=True,
+        blank=True
+    )
+
     summary = models.TextField(null=True, blank=True)
     description = MarkdownxField(null=True, blank=True)
 
@@ -516,6 +650,20 @@ class Episode(models.Model):
 
         return ''
 
+    def get_next_episode(self):
+        return self.podcast.episodes.exclude(
+            pk=self.pk
+        ).filter(
+            published__gte=self.published
+        ).order_by('published').first()
+
+    def get_previous_episode(self):
+        return self.podcast.episodes.exclude(
+            pk=self.pk
+        ).filter(
+            published__lte=self.published
+        ).order_by('published').last()
+
     class Meta:
         unique_together = ('guid', 'podcast')
         ordering = ('-published',)
@@ -639,3 +787,66 @@ class Page(models.Model):
     class Meta:
         unique_together = ('slug', 'podcast')
         ordering = ('ordering',)
+
+
+class Review(models.Model):
+    podcast = models.ForeignKey(
+        'Podcast',
+        related_name='reviews',
+        on_delete=models.CASCADE
+    )
+
+    directory = models.ForeignKey(
+        Directory,
+        on_delete=models.CASCADE,
+        related_name='reviews'
+    )
+
+    remote_id = models.CharField(max_length=255, editable=False)
+    country = models.CharField(max_length=50, null=True)
+    title = models.CharField(max_length=255)
+    body = models.TextField(null=True, blank=True)
+    author = models.CharField(max_length=50)
+    published = models.DateTimeField()
+    rating = models.DecimalField(decimal_places=1, max_digits=2, null=True)
+    approved = models.BooleanField(null=True)
+
+    def __str__(self):
+        return self.title
+
+    @property
+    def hash(self):
+        return md5(
+            (
+                '%s/%s' % (
+                    self.directory.name,
+                    self.author
+                )
+            ).encode('utf-8')
+        ).hexdigest()
+
+    def save(self, *args, **kwargs):
+        if self.approved is None:
+            self.approved = self.rating > 3
+        else:
+            self.approved = False
+
+        super().save(*args, **kwargs)
+
+    class Meta:
+        ordering = ('-published',)
+        get_latest_by = 'published'
+        unique_together = ('remote_id', 'directory', 'podcast')
+
+
+@transaction.atomic
+@receiver(websub.signals.published)
+def websub_published(sender, topic, data, *args, **kwargs):
+    for podcast in Podcast.objects.filter(
+        rss_feed_url=topic
+    ).select_for_update():
+        if data:
+            feed_data = parse_feed(data)
+            podcast.update_feed(feed_data, episode_callback=print)
+        else:
+            podcast.check_feed(episode_callback=print)
